/*
 * Copyright (c) 2020 QMTH All Rights Reserved.
 * Created by Deason on 2020-08-12 16:19:32
 */

package cn.com.qmth.examcloud.core.oe.admin.service.impl;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.persistence.criteria.Predicate;

import org.apache.commons.collections.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.basic.api.UserCloudService;
import cn.com.qmth.examcloud.core.basic.api.request.GetUserReq;
import cn.com.qmth.examcloud.core.basic.api.response.GetUserResp;
import cn.com.qmth.examcloud.core.oe.admin.base.utils.Check;
import cn.com.qmth.examcloud.core.oe.admin.dao.ExportTaskRepo;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExportTaskEntity;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.ExportTaskStatus;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.ExportTaskType;
import cn.com.qmth.examcloud.core.oe.admin.service.ExportTaskService;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.exporttask.ExportTaskInfo;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.exporttask.ExportTaskListReq;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.exporttask.ExportTaskListResp;
import cn.com.qmth.examcloud.examwork.api.ExamCloudService;
import cn.com.qmth.examcloud.examwork.api.bean.ExamBean;
import cn.com.qmth.examcloud.examwork.api.request.GetExamMapsReq;
import cn.com.qmth.examcloud.examwork.api.response.GetExamMapsResp;
import cn.com.qmth.examcloud.support.filestorage.FileStorageUtil;

/**
 * 导出任务相关接口
 */
@Service
public class ExportTaskServiceImpl implements ExportTaskService {

    @Autowired
    private ExportTaskRepo exportTaskRepo;

    @Autowired
    private ExamCloudService examCloudService;

    @Autowired
    private UserCloudService userCloudService;
    
    
    @Override
    public ExportTaskEntity findExportTaskToDispose() {
    	return exportTaskRepo.findExportTaskToDispose();
    }

    @Override
    @Transactional
    public Long addExportTask(ExportTaskInfo info) {
        Check.isNull(info.getRootOrgId(), "学校ID不能为空！");
        Check.isNull(info.getType(), "导出任务类型不能为空！");
        Check.isNull(info.getStatus(), "导出任务状态不能为空！");

        ExportTaskEntity entity = new ExportTaskEntity();
        entity.setType(info.getType());
        entity.setStatus(info.getStatus());
        entity.setStatusMsg(info.getStatusMsg());
        entity.setRootOrgId(info.getRootOrgId());
        entity.setExamId(info.getExamId());
        entity.setFilePath(info.getFilePath());
        entity.setCreator(info.getCreator());
        entity.setExportParam(info.getJsonParams());
        exportTaskRepo.save(entity);
        return entity.getId();
    }

    @Transactional
    @Override
    public void updateExportTaskStatus(Long taskId, ExportTaskStatus status) {
        this.updateExportTaskStatus(taskId, status, null, null);
    }
    
    @Transactional
    @Override
    public void startExportTask(Long taskId) {
    	exportTaskRepo.startExportTask(taskId, ExportTaskStatus.EXPORTING, new Date());
    }
    
    @Transactional
    @Override
    public void endExportTask(Long taskId) {
    	exportTaskRepo.endExportTask(taskId,new Date());
    }
    
    @Transactional
    @Override
    public void updateExportTaskStatus(Long taskId, ExportTaskStatus status, String statusMsg) {
        this.updateExportTaskStatus(taskId, status, statusMsg, null);
    }

    @Override
    @Transactional
    public void updateExportTaskStatus(Long taskId, ExportTaskStatus status, String statusMsg, String filePath) {
        Check.isNull(taskId, "导出任务ID不能为空！");
        Check.isNull(status, "导出任务状态不能为空！");

        exportTaskRepo.updateStatusAndStatusMsgById(taskId, status, statusMsg, filePath, new Date());
    }

    @Override
    @Transactional
    public void deleteExportTaskById(Long taskId) {
        Optional<ExportTaskEntity> optional = exportTaskRepo.findById(taskId);
        if (!optional.isPresent()) {
            return;
        }

        if (ExportTaskStatus.EXPORTING == optional.get().getStatus()) {
            throw new StatusException("500S05", "删除失败，当前导出任务尚未完成！");
        }

        exportTaskRepo.delete(optional.get());
    }

    @Override
    public boolean existUnFinishTask(ExportTaskType type, Long rootOrgId, Long examId) {
        Specification<ExportTaskEntity> spec = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();
            if (rootOrgId != null) {
                predicates.add(cb.equal(root.get("rootOrgId"), rootOrgId));
            }
            if (examId != null) {
                predicates.add(cb.equal(root.get("examId"), examId));
            }
            predicates.add(cb.equal(root.get("type"), type));
            predicates.add(cb.equal(root.get("status"), ExportTaskStatus.EXPORTING));
            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        List<ExportTaskEntity> tasks = exportTaskRepo.findAll(spec);
        if (CollectionUtils.isEmpty(tasks)) {
            return false;
        }

        boolean exist = false;

        // 超时情况更新为“失败”状态
        List<ExportTaskEntity> updateTasks = new ArrayList<>();
        for (ExportTaskEntity task : tasks) {
            Calendar calendar = Calendar.getInstance();
            calendar.setTime(task.getCreationTime());
            calendar.add(Calendar.MINUTE, 10);

            // 默认：10分钟
            if (new Date().before(calendar.getTime())) {
                exist = true;
                continue;
            }

            task.setStatus(ExportTaskStatus.ERROR);
            task.setStatusMsg("任务超时终止");
            updateTasks.add(task);
        }

        if (CollectionUtils.isNotEmpty(updateTasks)) {
            exportTaskRepo.saveAll(updateTasks);
        }

        return exist;
    }

    @Override
    public Page<ExportTaskListResp> getExportTaskList(ExportTaskListReq req, Long userId) {
        Check.isNull(req.getRootOrgId(), "学校ID不能为空！");

        if (req.getPageNo() == null || req.getPageNo() < 1) {
            req.setPageNo(1);
        }

        if (req.getPageSize() == null || req.getPageSize() < 1) {
            req.setPageSize(10);
        }

        Specification<ExportTaskEntity> spec = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();
            predicates.add(cb.equal(root.get("rootOrgId"), req.getRootOrgId()));
            predicates.add(cb.equal(root.get("creator"), userId));

            if (req.getExamId() != null) {
                predicates.add(cb.equal(root.get("examId"), req.getExamId()));
            }

            if (req.getType() != null) {
                predicates.add(cb.equal(root.get("type"), req.getType()));
            }

            if (req.getStatus() != null) {
                predicates.add(cb.equal(root.get("status"), req.getStatus()));
            }

            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        Sort sort = Sort.by(Sort.Order.desc("id"));
        Pageable pageable = PageRequest.of(req.getPageNo() - 1, req.getPageSize(), sort);
        Page<ExportTaskEntity> page = exportTaskRepo.findAll(spec, pageable);

        if (!page.hasContent()) {
            return new PageImpl<>(new ArrayList<>(), pageable, 0);
        }

        Set<Long> examIds = page.getContent().stream().filter(e -> e.getExamId() != null).map(e -> e.getExamId()).collect(Collectors.toSet());
        Map<Long, ExamBean> examMaps;
        if (CollectionUtils.isNotEmpty(examIds)) {
            GetExamMapsReq examsReq = new GetExamMapsReq();
            examsReq.setExamIds(examIds);
            GetExamMapsResp examsResp = examCloudService.getExamMaps(examsReq);
            examMaps = examsResp.getExamMaps();
        } else {
            examMaps = new HashMap<>();
        }

        GetUserReq reqUser = new GetUserReq();
        reqUser.setRootOrgId(req.getRootOrgId());
        reqUser.setUserId(userId);
        GetUserResp namesResp = userCloudService.getUser(reqUser);
        String createName = namesResp.getUserBean().getName();

        List<ExportTaskListResp> list = page.getContent()
                .stream().map(entity -> ofExportTask(entity, examMaps, createName)).collect(Collectors.toList());
        return new PageImpl<>(list, pageable, page.getTotalElements());
    }

    private ExportTaskListResp ofExportTask(ExportTaskEntity entity, Map<Long, ExamBean> examMaps, String createName) {
        ExportTaskListResp info = new ExportTaskListResp();

        info.setId(entity.getId());
        info.setType(entity.getType());
        info.setTypeTitle(entity.getType().getDesc());
        info.setStatus(entity.getStatus());
        info.setStatusTitle(entity.getStatus().getDesc());
        info.setStatusMsg(entity.getStatusMsg());
        info.setExamId(entity.getExamId());
        if (entity.getExamId() != null) {
            ExamBean exam = examMaps.get(entity.getExamId());
            info.setExamName(exam != null ? exam.getName() : "");
        }
        info.setFilePath(FileStorageUtil.realPath(entity.getFilePath()));
        info.setCreationTime(entity.getCreationTime());
        info.setUpdateTime(entity.getUpdateTime());
        info.setCreateName(createName);
        info.setStartTime(entity.getStartTime());
        info.setEndTime(entity.getEndTime());

        return info;
    }

}
